#-----------------------------------------------------------------------------
# Name        :   Add or Delete Control Points
# Description :   Add or Delete Control Points to a (non)rational Bezier Curve
# Context Menu:   Add or Delete Control Points
# Usage       :   Select an existing nonrational or rational Bezier curve and
#                 use Add or Delete Control Points to activate tool.
#             :   Then either click on a control point to delete the control
#                 point or click elsewhere to add a control point at that point. 
# Date        :   12/27/2014
# Type        :   Tool
#-----------------------------------------------------------------------------

# First we pull in the standard API hooks.
require 'sketchup.rb'
Sketchup::require 'jps_beziercurvemanager/ManipDirectly'

#-----------------------------------------------------------------------------
module JPS::BezierCurveManager

#-----------------------------------------------------------------------------
# Define the tool class for adding and deleting control points
class JPS_AddDelCtrlPntsBezierCurveTool < JPS_ManipulateBezierCurveTool

  def activate()
    super()
    Sketchup.status_text = "Click on a control point to delete it or elsewhere to add one at that location"
  end  

  def pick_point_to_move(x, y, view)
    old_pt_to_move = @pt_to_move
    super(x, y, view)
    if (@pt_to_move && @selection &&
        (@selection < 0 || @t != 0.0 || @q_selected))
      @pt_to_move = nil
      @selection = nil
      @t = 0.0
      @q_selected = false
    end
    old_pt_to_move != @pt_to_move
  end

  def find_index_to_insert_at()
    min_distance = @ip.position.distance @pts[0]
    index_to_insert_at = 0

    for i in 1..@degree
      distance = @ip.position.distance(Geom::Point3d.linear_combination(0.5, @pts[i-1], 0.5, @pts[i]))
      if(distance < min_distance)
        min_distance = distance
        index_to_insert_at = i
      end
    end

    if ( (@ip.position.distance @pts[@degree]) < min_distance)
      index_to_insert_at = @degree+1
    end

    return index_to_insert_at
  end


  def validate_weights()
    @weights[0] = @weights[@degree] = 1.0
    for i in 1..@degree-1
      if(@weights[i] < 0.1*@weights[i-1])
        @weights[i] = 0.1*@weights[i-1]
      elsif(@weights[i] > 10.0*@weights[i-1])
        @weights[i] = 10.0*@weights[i-1]
      end
    end
  end

  def onLButtonDown(flags, x, y, view)
    if(@pt_to_move)
      @pt_to_move = nil
      if (@degree == 1)
        UI.messagebox "Minimum degree is 1"
        view.invalidate
      else
        @pts.delete_at(@selection)
        @weights.delete_at(@selection)
        @degree -= 1
        validate_weights()
        @crvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
        calculate_qs()
        @state = 1
        @selection = nil
      end
      return
    end    

    if(@degree == 20)
      UI.messagebox("Maximum degree is 20.\nPlease consider\n" +
                    "1) removing redundancies by using the Change Degree tool or\n" +
                    "2) deleteing unnecessary control points using this tool or\n" +
                    "3) subdividing the curve using the Subdivide tool.")
      return
    end

    @ip.pick view, x, y, Sketchup::InputPoint.new(@pts[0])
    if (@ip.valid?)
      insert_index = find_index_to_insert_at()
      @pts.insert(insert_index, @ip.position)
      @weights.insert(insert_index, 1.0)
      @degree += 1
      @crvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
      calculate_qs()
      @state = 1
    end
  end

  def draw(view)
    user_did_undo = @user_did_undo
    if (!user_did_undo)
      qs_orig = @qs
      @qs = nil
    end
    super(view)
    if (!user_did_undo)
      @qs = qs_orig
    end
  end

end # class JPS_AddDelCtrlPntsBezierCurveTool
end #module JPS::BezierCurveManager