#-----------------------------------------------------------------------------
# Name        :   Change Degree of Bezier Curve
# Description :   Increase or decrease the number of control points
# Context Menu:   Change Degree
# Usage       :   Select an existing nonrational or rational Bezier curve and
#                 use Change Degree to activate the tool.
#             :   Then enter the new degree greater than 0 and less than 21 in
#                 the Value Control Box(VCB). Degree elevation will not affect
#                 the shape of the curve but degree reduction may.  
# Date        :   12/19/2014
# Type        :   Tool
#-----------------------------------------------------------------------------

# First we pull in the standard API hooks.
require 'sketchup.rb'
Sketchup::require 'jps_beziercurvemanager/ManipDirectly'

#-----------------------------------------------------------------------------
module JPS::BezierCurveManager

#-----------------------------------------------------------------------------
# Define the tool class for changing the degree of Bezier curves.
class JPS_ChangeBezierCurveDegreeTool < JPS_ManipulateBezierCurveTool

  def activate
    super()
    Sketchup::set_status_text "Type the new degree for the selected Bezier Curve and press Enter or Return", SB_PROMPT
    Sketchup::set_status_text "Degree", SB_VCB_LABEL
    Sketchup::set_status_text @degree, SB_VCB_VALUE
    @allow_change_status_text_on_mouse_move = false
  end

  # For this tool, allow vcb text entry while the tool is active.
  def enableVCB?
    return true
  end

  # This is based on the technique described in "CAGD  A Practical Guide, 5th
  # Edition" by Gerald Farin. page 236
  def increaseDegreeByOne()
    newpts = Array.new(@degree+2){ Geom::Point3d.new }
    newweights = Array.new(@degree+2){ 0.0 }
    newpts[0] = @pts[0]
    newweights[0] = 1.0
    for i in 1..@degree
      alpha = i.to_f/(@degree+1)
      newweights[i] = @weights[i-1]*alpha + @weights[i]*(1.0 - alpha)
      newpts[i] = Geom::Point3d.linear_combination((@weights[i-1]*alpha)/newweights[i], @pts[i-1], (@weights[i]*(1.0-alpha))/newweights[i], @pts[i])
    end
    newpts[@degree+1] = @pts[@degree]
    newweights[@degree+1] = 1.0
    @pts = newpts
    @weights = newweights
    @degree += 1
  end
  
  def elevateDegree(newdegree)
    while @degree < newdegree
      increaseDegreeByOne()  
    end 
  end

  # This is based on the technique described in "CAGD  A Practical Guide, 5th
  # Edition" by Gerald Farin. page 85
  def findNewPoints()

    m = Array.new(@degree+1){Array.new(@degree) {0.0} }
    m[0][0] = 1.0
    for i in 1..@degree-1
      alpha = i.to_f/@degree
      m[i][i-1] = @weights[i-1]*alpha
      m[i][i] = @weights[i]*(1.0 - alpha)
      den = m[i][i-1] + m[i][i]
      m[i][i-1] /= den
      m[i][i] /= den
    end
    m[@degree][@degree-1] = 1.0

    mtm = Array.new(@degree){Array.new(@degree) {0.0} }
    mtm[0][0] = 1.0 + m[1][0]*m[1][0] 
    mtm[0][1] = m[1][0]*m[1][1]
    for i in 1...@degree-1
      mtm[i][i-1] = mtm[i-1][i]
      for j in i..i+1
        for k in i-1..i+1
          mtm[i][j] += m[k][i]*m[k][j]
        end
      end
    end
    mtm[@degree-1][@degree-2] = m[@degree-1][@degree-2]*m[@degree-1][@degree-1]
    mtm[@degree-1][@degree-1] = 1+ m[@degree-1][@degree-1]*m[@degree-1][@degree-1]

    l = Array.new(@degree){Array.new(@degree) {0.0} }
    l[0][0] = Math.sqrt(mtm[0][0])
    for i in 1..@degree-1
      l[i][i-1] = mtm[i][i-1]/l[i-1][i-1]
      l[i][i] = Math.sqrt(mtm[i][i] - l[i][i-1]*l[i][i-1])
    end

    mtb1 = Array.new(@degree){ Geom::Point3d.new }
    for i in 0..@degree-1
      for j in 0..@degree
        for k in 0..2
          mtb1[i][k] += m[j][i]*@pts[j][k]
        end
      end
    end

    #Forward substitution
    y = Array.new(@degree){ Geom::Point3d.new }
    for i in 0..@degree-1
      for k in 0..2
        for j in 0..i-1
          y[i][k] += l[i][j]*y[j][k]
        end
        y[i][k] = (mtb1[i][k] - y[i][k])/l[i][i]
      end
    end

    #Backward substitution
    newpts = Array.new(@degree){ Geom::Point3d.new }
    (@degree-1).downto(0) { |i|
      for k in 0..2
        for j in i+1..@degree-1
          newpts[i][k] += l[j][i]*newpts[j][k]
        end
        newpts[i][k] = (y[i][k] - newpts[i][k])/l[i][i]
      end
    }

    @pts = newpts
  end

  
  def findNewWeights()
     
    mweights = Array.new(@degree+1){Array.new(@degree) {0.0} }
    mweights[0][0] = 1.0
    for i in 1..@degree-1
      alpha = i.to_f/@degree
      mweights[i][i-1] = alpha
      mweights[i][i] = (1.0 - alpha)
    end
    mweights[@degree][@degree-1] = 1.0

    mtmweights = Array.new(@degree){Array.new(@degree) {0.0} }
    mtmweights[0][0] = 1.0 + mweights[1][0]*mweights[1][0]
    mtmweights[0][1] = mweights[1][0]*mweights[1][1]
    for i in 1...@degree-1
      mtmweights[i][i-1] = mtmweights[i-1][i]
      for j in i..i+1
        for k in i-1..i+1
          mtmweights[i][j] += mweights[k][i]*mweights[k][j]
        end
      end
    end
    mtmweights[@degree-1][@degree-2] = mweights[@degree-1][@degree-2]*mweights[@degree-1][@degree-1]
    mtmweights[@degree-1][@degree-1] = 1+ mweights[@degree-1][@degree-1]*mweights[@degree-1][@degree-1]


    lweights = Array.new(@degree){Array.new(@degree) {0.0} }
    lweights[0][0] = Math.sqrt(mtmweights[0][0])
    for i in 1..@degree-1
      lweights[i][i-1] = mtmweights[i][i-1]/lweights[i-1][i-1]
      lweights[i][i] = Math.sqrt(mtmweights[i][i] - lweights[i][i-1]*lweights[i][i-1])
    end

    mtb1weights = Array.new(@degree){ 0.0 }
    for i in 0..@degree-1
      for j in 0..@degree
        mtb1weights[i] += mweights[j][i]*@weights[j]
      end
    end

    #Forward substitution
    yweights = Array.new(@degree){ 0.0 }
    for i in 0..@degree-1
      for j in 0..i-1
        yweights[i] += lweights[i][j]*yweights[j]
      end
      yweights[i] = (mtb1weights[i] - yweights[i])/lweights[i][i]
    end

    #Backward substitution
    newweights = Array.new(@degree){ 0.0 }
    (@degree-1).downto(0) { |i|
      for j in i+1..@degree-1
        newweights[i] += lweights[j][i]*newweights[j]
      end
      newweights[i] = (yweights[i] - newweights[i])/lweights[i][i]
    }

    @weights = newweights
  end

  def reduceDegreeByOne()
    findNewWeights()
    findNewPoints()
    @degree -= 1  
  end

  # This is based on the technique described in "CAGD  A Practical Guide, 5th
  # Edition" by Gerald Farin. page 235
  def reparameterize()    
    if (@degree > 1)
      c = (@weights[0]/@weights[@degree])**(1.0/@degree)
    end
    
    ci = 1.0
    for i in 1..@degree-1
      ci *= c
      @weights[i] *= ci/@weights[0]
    end
   
    @weights[0] = @weights[@degree] = 1.0
  end

  def stabilize_weights()
    
    reparameterize()
    
    for i in 1..@degree-1
      if(@weights[i] < 0.1)
        @weights[i] = 0.1
      elsif(@weights[i] > 10.0)
        @weights[i] = 10.0
      end
      if(@weights[i] < 0.1*@weights[i-1])
        @weights[i] = 0.1*@weights[i-1]
      elsif(@weights[i] > 10.0*@weights[i-1])
        @weights[i] = 10.0*weights[i-1]
      end
    end
    
  end

  def reduceDegree(newdegree)
    while newdegree < @degree
      reduceDegreeByOne()  
    end 
    stabilize_weights()
  end

  def onUserText(text, view)
    # get the degree from the text
    newdegree = text.to_i
    if(newdegree < 1)
      UI.messagebox "Minimum degree is 1"
      Sketchup::set_status_text @degree, SB_VCB_VALUE
    elsif(newdegree > 20)
      UI.messagebox("Maximum degree is 20.\nPlease consider\n" +
                    "1) removing redundancies by reducung the degree using this tool or\n" +
                    "2) deleteing control points using the Add Delete Control Points tool or\n" +
                    "3) subdividing the curve using the Subdivide tool.")
      Sketchup::set_status_text @degree, SB_VCB_VALUE
    elsif(newdegree != @degree)
      if ( newdegree > @degree )
        elevateDegree(newdegree)
      else
        reduceDegree(newdegree)
      end
      @crvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
      calculate_qs()
      @selection = nil
      @state = 1
      onLButtonUp(0, 0, 0, view)
    else 
      UI.beep
      Sketchup::set_status_text @degree, SB_VCB_VALUE
    end
  end


end # class JPS_ChangeBezierCurveDegreeTool
end #module JPS::BezierCurveManager