#-----------------------------------------------------------------------------
# Name        :   Edit using weight points
# Description :   A tool to add "weights' to Bezier Curve control points
# Menu Item   :   None
# Context Menu:   Edit using weight points
# Usage       :   Select a (Rational) Bezier curve, arc, circle or line and 
#                 use Edit using weight points to activate the tool.
#                 The grab a weight point, marked by triangles, and move it
#                 along its control polygon line to influence the shape of the
#                 curve. It may be preferable for a designer to use these 
#                 geometric handles rather than requiring to input numbers 
#                 for the weights.
# Date        :   02/08/2015
# Type        :   Tool
#-----------------------------------------------------------------------------

require 'sketchup.rb'
require 'jps_beziercurvemanager/su_bezier.rb'

#-----------------------------------------------------------------------------
module JPS::BezierCurveManager

#-----------------------------------------------------------------------------
# Define the tool class for Rational Bezier Curves.
# This is based on the technique described in "CAGD  A Practical Guide, 5th
# Edition" by Gerald Farin. page 227

class JPS_ManipulateBezierCurveTool < EditBezierTool

  def is_tool_active()
    return @tool_active
  end 
   
  def calculate_qs()
    @qs = []
    for i in 0...@degree-1
      @qs[i] = Geom::Point3d.linear_combination @weights[i]/(@weights[i]+@weights[i+1]), @pts[i], @weights[i+1]/(@weights[i]+@weights[i+1]), @pts[i+1]
    end
  end

  def get_pts_and_weights_from_arccurve()

   theta = @curve.end_angle - @curve.start_angle

   if (theta < 2.0*Math.acos(0.1)) #168.52165904546642737497018219214 degrees
     @pts = Array.new(3, Geom::Point3d.new)
     @weights = Array.new(3, 1.0)
     @weights[1] = Math.cos(theta/2)
     @pts[0] = @vertices[0].position
     @pts[2] = @vertices[@numseg].position #pt1 needs pt2
     @pts[1] = Geom::Point3d.linear_combination(1.0, @curve.center, 1.0/(@weights[1]*@weights[1]),
                 Geom::Point3d.linear_combination(-1.0, @curve.center, 0.5, 
                   Geom::Point3d.linear_combination(1.0, @pts[0], 1.0, @pts[2])))
   elsif (theta < 2.0*Math.acos(-0.35)) #220.97463022944532693351316331888 degrees
     @pts = Array.new(4, Geom::Point3d.new)
     @weights = Array.new(4, 1.0)
     @weights[1] = @weights[2] =(1.0 + 2.0*Math.cos(theta/2))/3.0
     @pts[0] = @vertices[0].position
     @pts[3] = @vertices[@numseg].position #pt2 needs pt3
     e = 2.0*Math.sin(theta/2.0)/(1.0 + 2.0*Math.cos(theta/2.0))
     @pts[1] = Geom::Point3d.linear_combination(1.0, @pts[0], e, Geom::Point3d.new(@curve.normal.cross(@pts[0] - @curve.center).to_a))
     @pts[2] = Geom::Point3d.linear_combination(1.0, @pts[3], e, Geom::Point3d.new((@pts[3] - @curve.center).cross(@curve.normal).to_a))
   elsif (theta < 4.0*Math.acos(0.1)) #337.04331809093285474994036438428 degrees
     @weights = Array.new(5, 1.0)
     c = Math.cos(theta/4.0)
     @weights[1] = @weights[3] = c
     @weights[2] = (1.0 + 2.0*c*c)/3.0
     @pts = Array.new(5, Geom::Point3d.new)
     @pts[0] = @vertices[0].position
     for i in 1..3
       @pts[i] = Geom::Point3d.linear_combination(1.0, @curve.center, 1.0/@weights[i], Geom::Point3d.new(Geom::Vector3d.linear_combination(Math.cos(i*theta/4.0), @curve.xaxis, Math.sin(i*theta/4.0), @curve.yaxis).to_a))
     end
     @pts[4] = @vertices[@numseg].position
   elsif (true || theta < 2.0*Math::PI) #Reduce degree has a problem with the degree 5 representation of a circle below. Hence the true.
     @weights = Array.new(7, 1.0)
     c = Math.cos(theta/6.0)
     @weights[1] = @weights[5] = c
     @weights[2] = @weights[4] = (1.0 + 4.0*c*c)/5.0
     @weights[3] = c*(3.0 + 2.0*c*c)/5.0
     @pts = Array.new(7, Geom::Point3d.new)
     @pts[0] = @vertices[0].position
     for i in 1..5
       @pts[i] = Geom::Point3d.linear_combination(1.0, @curve.center, 1.0/@weights[i], Geom::Point3d.new(Geom::Vector3d.linear_combination(Math.cos(i*theta/6.0), @curve.xaxis, Math.sin(i*theta/6.0), @curve.yaxis).to_a))
     end
     @pts[6] = @vertices[@numseg].position
     @degree = 6
     #reduceDegreeByOne() not PERFECT
   else
     @pts = Array.new(6, Geom::Point3d.new)
     @pts[0] = @vertices[0].position
     @pts[1] = Geom::Point3d.linear_combination(1.0, @curve.center, 1.0, Geom::Point3d.new(Geom::Vector3d.linear_combination(1.0, @curve.xaxis, 4.0, @curve.yaxis).to_a))
     @pts[2] = Geom::Point3d.linear_combination(1.0, @curve.center, 1.0, Geom::Point3d.new(Geom::Vector3d.linear_combination(-3.0, @curve.xaxis, 2.0, @curve.yaxis).to_a))
     @pts[3] = Geom::Point3d.linear_combination(1.0, @curve.center, 1.0, Geom::Point3d.new(Geom::Vector3d.linear_combination(-3.0, @curve.xaxis, -2.0, @curve.yaxis).to_a))
     @pts[4] = Geom::Point3d.linear_combination(1.0, @curve.center, 1.0, Geom::Point3d.new(Geom::Vector3d.linear_combination(1.0, @curve.xaxis, -4.0, @curve.yaxis).to_a))
     @pts[5] = @vertices[@numseg].position
     @weights = [1.0, 0.2, 0.2, 0.2, 0.2, 1.0]
   end
  end


  def activate
    super()
    # Make sure that there is really a Bezier curve selected
    if( not @curve )
      return
    end

    if( not @pts )
      if( @curve.kind_of? Sketchup::Edge )
        @pts = Array.new(2, Geom::Point3d.new)
        @pts[0] = @curve.start.position
        @pts[1] = @curve.end.position
        @weights = Array.new(2, 1.0)
      else
        get_pts_and_weights_from_arccurve()
      end 
      if( not @pts )
        UI.beep
        Sketchup.active_model.select_tool nil
        return
      end

      # Get the curve points from the vertices
      @vertices = @curve.vertices
      @crvpts = @vertices.collect {|v| v.position}
      if( @curve.kind_of? Sketchup::Edge )
        @numseg = 20
      else
        @numseg = @vertices.length - 1
      end
    end

    @degree = @pts.length-1
    calculate_qs()
    @tool_active = true
    Sketchup.active_model.active_view.invalidate
    Sketchup.status_text = "Change the location of weight points, marked by triangles, to recompute the weights and influence the shape of the curve"
    @allow_change_status_text_on_mouse_move = true
  end

  # For this tool, allow vcb text entry while the tool is active.
  def enableVCB?
    return true
  end

  def onUserText(text, view)
    if (@selection && @selection >= 0 && (@t <= 0.01 || @t >= 0.99))
      # get the weight from the text
      newweight = text.to_f
      if(@selection == 0 || @selection == @degree)
        UI.messagebox "Lets leave the weight of the end control points at 1.0 in standard form."
      elsif(newweight < 0.1)
        UI.messagebox "Minimum weight is 0.1 for numerical stability.\n" +
                      "Please consider\n" +
                      "1) increasing the weight of the other control points or\n" +
                      "2) deleting this control point."
        Sketchup::set_status_text @weights[@selection], SB_VCB_VALUE
      elsif(newweight > 10.0)
        UI.messagebox "Maximimum weight is 10.0 for numerical stability.\n" +
                      "Please consider\n" +
                      "1) decreasing the weight of the other control points or\n" +
                      "2) adding another control point close to this one." 
        Sketchup::set_status_text @weights[@selection], SB_VCB_VALUE
      elsif(newweight < 0.1*@weights[@selection-1] || newweight < 0.1*@weights[@selection+1])
        UI.messagebox "Minimum weight is one-tenth of adjacent weights for numeric stability.\n" +
                      "Please consider\n" +
                      "1) decreasing weight(s) of offending adjacent control point(s) first or\n" +
                      "2) deleting this control point."
        Sketchup::set_status_text @weights[@selection], SB_VCB_VALUE
      elsif(newweight > 10.0*@weights[@selection-1] || newweight > 10.0*@weights[@selection+1])
        UI.messagebox "Maximimum weight is ten times adjacent weights for numeric stability.\n" +
                      "Please consider\n" +
                      "1) increasing weight(s) of offending adjacent control point(s) first or\n" +
                      "2) adding another control point close to this one."
        Sketchup::set_status_text @weights[@selection], SB_VCB_VALUE
      else
        @weights[@selection] = newweight
        @crvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
        calculate_qs()
        @state = 1
        onLButtonUp(0, 0, 0, view)
      end
    else 
      UI.messagebox "Hover over the control point you want to change the weight of."
      Sketchup::set_status_text "", SB_VCB_VALUE
    end
  end

  def resume(view)
    activate
    super(view)
  end

  def deactivate(view)
    super(view)
    @tool_active = false
  end

  def suspend(view)
    deactivate(view)
  end

  def pick_point_to_move(x, y, view)
    old_pt_to_move = @pt_to_move
    @q_selected = false
    @t = 0.0
    ph = view.pick_helper x, y
    if (@degree == 1)
      @selection = nil
    elsif (@degree == 2)
      if (ph.test_point @qs[0])
        @selection = 0;
      else
        @selection = nil
      end
    else
      @selection = ph.pick_segment @qs
    end
    if (@selection && @selection >= 0)
      @pt_to_move = @qs[@selection]
      @q_selected = true
      @q_move_segment = [Geom::Point3d.linear_combination(10.0/11.0, @pts[@selection], 1.0/11.0, @pts[@selection+1]),
                         Geom::Point3d.linear_combination(1.0/11.0, @pts[@selection], 10.0/11.0, @pts[@selection+1])]
    else
      super(x, y, view)
      if(@selection && @selection < 0 && @degree == 1 && @subdivide_tool_is_active)
        @pt_to_move = nil #Neglect points on the control line of a line so that we can select point on the line itself to subdivide at
      end
    end
    if (@pt_to_move == nil)
      @selection = ph.pick_segment @crvpts
      if( @selection )
        if( @selection < 0 )
          # We got a point on the curve.  Compute the point closest
          # to the pick ray.
          pickray = view.pickray x, y
          i = -@selection
          segment = [@crvpts[i-1], @crvpts[i]]
          result = Geom.closest_points segment, pickray
          @pt_to_move = result[0]
          @t = ((i-1) + ((@crvpts[i-1].distance @pt_to_move)/(@crvpts[i-1].distance @crvpts[i])))/(@numseg)
        else
          # we got an intersection point
          @t = @selection.to_f/(@numseg)
          @pt_to_move = @crvpts[@selection]
        end
        if (@t <= 0.01 || @t >= 0.99)
          @pt_to_move = nil
          @selection = nil
          @t = 0.0
        end
      end
    end
    old_pt_to_move != @pt_to_move
  end

  def choose( n, r )
    if (r < 0)
      return 0
    elsif (r == 0)
      return 1
    end

    num = n
    den = r
    
    for i in 1..(r-1)
      num *= (n-i)
      den *= i
    end
    
    return (num/den)
  end  

  def bern( n, i, u)
    return (choose(n, i)*(u ** i)*((1.0-u)**(n-i)))
  end 

  def onLButtonDown(flags, x, y, view)
    super(flags, x, y, view)
    if( @state == 1 && @t > 0.01 && @t < 0.99)
      @aArr = []
      @a = 0.0
      den = 0.0
      for i in 0..@degree
        @aArr[i] = @weights[i]*bern(@degree, i, @t)
        den += @aArr[i]
      end
      for i in 1..@degree-1
        @aArr[i] /= den
        @a += @aArr[i]*@aArr[i]
      end
    end
  end

  def convert_arccurve_or_edge_to_bezcurve()
    if (@curve.kind_of?(Sketchup::ArcCurve) || @curve.kind_of?(Sketchup::Edge))
      # delete the edge/arc
      if (@curve.kind_of?(Sketchup::Edge))
        @curve.erase!
      else
        @curve.each_edge { |e| e.erase! }
      end
        
      # create the new curve
      edges = Sketchup.active_model.active_entities.add_curve(@crvpts);
      @curve = edges[0].curve
    
      # Update the control points stored with the curve
      @curve.set_attribute "jps", "crvtype", "Rational Bezier"
      @curve.set_attribute "jps", "crvpts", @pts
      @curve.set_attribute "jps", "weights",  @weights  

      @vertices = @curve.vertices
      @numseg = @vertices.length - 1
    end
  end

  def onLButtonUp(flags, x, y, view)
    return if not @state == 1

    if (@curve.kind_of?(Sketchup::ArcCurve) || @curve.kind_of?(Sketchup::Edge))
      @crvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
      model = Sketchup.active_model
      model.selection.clear
    
      model.start_operation "Edit Rational Bezier Curve"
        edges = convert_arccurve_or_edge_to_bezcurve()  
      model.commit_operation
      model.selection.add @curve.edges 
    else
      super(flags, x, y, view)
    end

    @state = 0
  end  
   
  def recompute_weights()
    for i in @selection+1..@degree-1
      @weights[i] = (@qs[i-1].distance(@pts[i-1])/@qs[i-1].distance(@pts[i]))*@weights[i-1]
      if(@weights[i] < 0.1)
        @weights[i] = 0.1
      elsif(@weights[i] > 10.0)
        @weights[i] = 10.0
      end
    end
  end

  def onMouseMove(flags, x, y, view)
    # Make sure that the control polygon is shown
    view.invalidate if not @drawn
    @ip.clear
    
    # Move the selected point if state = 1
    if(@state == 1 && @selection && (@q_selected || (@t > 0.01 && @t < 0.99)))
      if(@q_selected)
        # Moving a weight point
        # This is based on the technique described in "CAGD  A Practical Guide,
        # 5th Edition" by Gerald Farin. page 230
        ph = view.pick_helper x, y
        new_q = ph.pick_segment @q_move_segment
        if (new_q)
          if (new_q < 0)
            # We got a point on the control line.  Compute the point closest
            # to the pick ray.
            result = Geom.closest_points(@q_move_segment, view.pickray(x, y))
            @qs[@selection] = result[0]
          else
            # we got an end point
            @qs[@selection] = @q_move_segment[new_q]
          end
          pt = @qs[@selection]
          recompute_weights()
        end
      else
        # Moving a point on the curve
        # This is based on the technique described in "CAGD  A Practical Guide,
        # 5th Edition" by Gerald Farin. page 112
        @ip.pick view, x, y
        return if not @ip.valid?
        pt = @ip.position
        vec = pt - @pt_to_move
        for i in 1..@degree-1
          for j in 0..2
            @pts[i][j] += (@aArr[i]/@a)*vec[j]
          end
        end
      end
      @pt_to_move = pt
      @crvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
      calculate_qs()
      view.invalidate
    else
      super(flags, x, y, view)
      if(@state == 1 && @selection && @ip.valid?)
        calculate_qs()
      end
    end
    if (@allow_change_status_text_on_mouse_move)
      if (@selection && @selection > 0 && @selection < @degree && !@q_selected && (@t <= 0.01 || @t >= 0.99))
        Sketchup::set_status_text "Weight", SB_VCB_LABEL
        Sketchup::set_status_text @weights[@selection], SB_VCB_VALUE
      else
        Sketchup::set_status_text "", SB_VCB_LABEL
        Sketchup::set_status_text "", SB_VCB_VALUE
      end
    end
  end

  def draw(view)
    
    if (@user_did_undo)
      @user_did_undo = false
      if (Sketchup.active_model.selection.count == 0)
        deactivate(view)
        Sketchup.active_model.select_tool nil
        return
      end
      activate()
    end
      
    pt_to_move_orig = @pt_to_move
    if (@selection && @selection >= 0)
      if(@q_selected)
        view.draw_points(@qs[@selection], 10, 7, "blue")
        @pt_to_move = nil
      elsif(@t <= 0.01 || @t >= 0.99)
        view.draw_points(@pts[@selection], 10, 2, "green")
      end
    end
    super(view)
    @pt_to_move = pt_to_move_orig
    view.draw_points(@pts, 10, 1, "green")
    if (@degree > 1 && @qs != nil)
      view.draw_points(@qs, 10, 6, "blue")
    end
    for i in 0..@degree
      view.draw_text view.screen_coords(@pts[i]), sprintf("%.2f", @weights[i])
    end
    if( @state == 1 && @ip.valid? ) 
      @ip.draw(view)
    end
  end

  # The reason can be one of the following values: 
  # 0: the user canceled the current operation by hitting the escape key.
  # 1: the user re-selected the same tool from the toolbar or menu.
  # 2: the user did an undo while the tool was active.
  def onCancel(reason, view)
    if(reason == 2)
      @user_did_undo = true 
    end
    view.invalidate
  end

end # class JPS_ManipulateBezierCurveTool
end # module JPS::BezierCurveManager
