#-----------------------------------------------------------------------------
# Name        :   Subdivide Bezier Curve Tool
# Description :   Subdivide the selected Bezier Curve at the chosen point
# Context Menu:   Subdivide
# Usage       :   Select an existing nonrational or rational Bezier curve, arc,
#                 circle or line and choose Subdivide to activate the tool.
#             :   Then choose the point on the curve to split the curve into
#                 two independent curves.  
# Date        :   02/08/2015
# Type        :   Tool
#-----------------------------------------------------------------------------

# First we pull in the standard API hooks.
require 'sketchup.rb'
Sketchup::require 'jps_beziercurvemanager/ManipDirectly'

#-----------------------------------------------------------------------------
module JPS::BezierCurveManager

#-----------------------------------------------------------------------------
# Define the tool class for subdivision of Bezier curves.
class JPS_SubDivideBezierCurveTool < JPS_ManipulateBezierCurveTool

  def activate()
    super()
    Sketchup.status_text = "Choose a point on the selected Bezier curve where you want to subdivide it"
    @subdivide_tool_is_active = true
  end
  
  # This is based on the technique described in "CAGD  A Practical Guide, 5th
  # Edition" by Gerald Farin. page 231
  def subdivide()
  
    @curve2pts = Array.new(@degree+1)
    @curve2weights = Array.new(@degree+1)
    @curve2pts[@degree] = @pts[@degree]
    @curve2weights[@degree] = @weights[@degree]

    oldpts = @pts
    oldweights = @weights
    for i in 1..@degree
      newpts = Array.new(@degree-i+1)
      newweights = Array.new(@degree-i+1)
      for j in 0..@degree-i
        newweights[j] = (1.0-@t)*oldweights[j] + @t*oldweights[j+1]
        newpts[j] = Geom::Point3d.linear_combination((1.0-@t)*oldweights[j]/newweights[j], oldpts[j], @t*oldweights[j+1]/newweights[j], oldpts[j+1])
      end
      @pts[i] = newpts[0]
      @weights[i] = newweights[0]
      @curve2pts[@degree-i] = newpts[@degree-i]
      @curve2weights[@degree-i] = newweights[@degree-i]
      oldpts = newpts
      oldweights = newweights
    end

  end  

  # This is based on the technique described in "CAGD  A Practical Guide, 5th
  # Edition" by Gerald Farin. page 235
  def reparameterize_curves()    
    if (@degree > 1)
      c = (@weights[0]/@weights[@degree])**(1.0/@degree)
      curve2c = (@curve2weights[0]/@curve2weights[@degree])**(1.0/@degree)
    end
    
    ci = curve2ci = 1.0
    for i in 1..@degree-1
      ci *= c
      curve2ci *= curve2c
      @weights[i] *= ci/@weights[0]
      @curve2weights[i] *= curve2ci/@curve2weights[0]
    end
   
    @weights[0] = @weights[@degree] = @curve2weights[0] = @curve2weights[@degree] = 1.0
  end

  def onLButtonDown(flags, x, y, view)
    # Don't do anything if a point on the curve is selected
    if( @state == 1 && @selection && @t > 0.01 && @t < 0.99)
    else
      super(flags, x, y, view)
    end
  end 

  def onMouseMove(flags, x, y, view)
    # Don't do anything if a point on the curve is selected
    if( @state == 1 && @selection && @t > 0.01 && @t < 0.99)
      # Make sure that the control polygon is shown
      view.invalidate if not @drawn
    else
      super(flags, x, y, view)
    end
  end

  def onLButtonUp(flags, x, y, view)  
    if( @state == 1 && @t > 0.01 && @t < 0.99)
      subdivide()
      reparameterize_curves()
      @curvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
      calculate_qs()
      curve2curvpts = JPS::BezierCurveManager.points(@curve2pts, @curve2weights, @numseg)
      model = view.model
      model.selection.clear
      model.start_operation "Subdivide Bezier Curve"
        if (@curve.kind_of?(Sketchup::Edge))
          @curve.erase!
        else
          @curve.each_edge { |e| e.erase! }
        end

        # create the 1st curve
        edges1 = model.active_entities.add_curve(@curvpts);
        curve = edges1[0].curve
        curve.set_attribute "jps", "crvtype", "Rational Bezier"
        curve.set_attribute "jps", "crvpts", @pts
        curve.set_attribute "jps", "weights",  @weights
         
        # create the 2nd curve
        edges = model.active_entities.add_curve(curve2curvpts);
        curve = edges[0].curve
        curve.set_attribute "jps", "crvtype", "Rational Bezier"
        curve.set_attribute "jps", "crvpts", @curve2pts
        curve.set_attribute "jps", "weights",  @curve2weights
      model.commit_operation
      model.selection.add edges1
      activate()
    else
      super(flags, x, y, view)
    end
  end

end # class JPS_SubDivideBezierCurveTool
end #module JPS::BezierCurveManager