# Copyright 2013, Trimble Navigation Limited

# This software is provided as an example of using the Ruby interface
# to SketchUp.

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

require 'sketchup.rb'

# Ruby implementation of Rational Bezier curves

module JPS::BezierCurveManager

# Evaluate a Rational Bezier curve at a parameter.
# The curve is defined by an array of its control points and
# another array of the points' corresponding "weights".
# The parameter ranges from 0 to 1
# This is based on the technique described in "CAGD  A Practical Guide, 4th
# Edition" by Gerald Farin. page 227

def self.eval(pts, weights, t)

    degree = pts.length - 1
    if degree < 1
        return nil
    end
    
    t1 = 1.0 - t
    fact = 1.0
    n_choose_i = 1

    x = weights[0]*pts[0].x * t1
    y = weights[0]*pts[0].y * t1
    z = weights[0]*pts[0].z * t1
    den = weights[0] * t1
    
    for i in 1...degree
        fact = fact*t
        n_choose_i = n_choose_i*(degree-i+1)/i
        fn = fact * n_choose_i
        x = (x + fn*weights[i]*pts[i].x) * t1
        y = (y + fn*weights[i]*pts[i].y) * t1
        z = (z + fn*weights[i]*pts[i].z) * t1
        den = (den + fn*weights[i]) * t1
    end

    x = x + fact*t*weights[degree]*pts[degree].x
    y = y + fact*t*weights[degree]*pts[degree].y
    z = z + fact*t*weights[degree]*pts[degree].z
    den = den + fact*t*weights[degree]

    Geom::Point3d.new(x/den, y/den, z/den)
    
end # method eval

# Evaluate the curve at a number of points and return the points in an array
def self.points(pts, weights, numpts)
    
    curvepts = []
    dt = 1.0 / numpts

    # evaluate the points on the curve
    for i in 0..numpts
        t = i * dt
        curvepts[i] = self.eval(pts, weights, t)
    end
    
    curvepts
end

#-----------------------------------------------------------------------------
# Define the tool class for editing Bezier curves

class EditBezierTool

def activate

    @state = 0
    @drawn = false
    @selection = nil
    @pt_to_move = nil
    
    # Make sure that there is really a Bezier curve selected
    @curve = JPS::BezierCurveManager.selected_curve
    if( not @curve )
        Sketchup.active_model.select_tool nil
        return
    end
    
    # Get the control points
    @pts = @curve.get_attribute "skp", "crvpts"
    if( not @pts )
     @pts = @curve.get_attribute "jps", "crvpts"
     if (@pts)
       @weights = @curve.get_attribute "jps", "weights"
     end
    else
      @weights = Array.new(@pts.length, 1.0)
    end
    
    # Get the curve points from the vertices
    @vertices = @curve.vertices
    @crvpts = @vertices.collect {|v| v.position}
    @numseg = @vertices.length - 1
    
    @ip = Sketchup::InputPoint.new
end

def deactivate(view)
    view.invalidate if @drawn
    @ip = nil
end

def resume(view)
    @drawn = false
end

def pick_point_to_move(x, y, view)
    old_pt_to_move = @pt_to_move
    ph = view.pick_helper x, y
    @selection = ph.pick_segment @pts
    if( @selection )
        if( @selection < 0 )
            # We got a point on a segment.  Compute the point closest
            # to the pick ray.
            pickray = view.pickray x, y
            i = -@selection
            segment = [@pts[i-1], @pts[i]]
            result = Geom.closest_points segment, pickray
            @pt_to_move = result[0]
        else
            # we got a control point
            @pt_to_move = @pts[@selection]
        end
    else
        @pt_to_move = nil
    end
    old_pt_to_move != @pt_to_move
end

def onLButtonDown(flags, x, y, view)
    # Select the segment or control point to move
    self.pick_point_to_move x, y, view
    @state = 1 if( @selection )
end

def onLButtonUp(flags, x, y, view)
    return if not @state == 1
    @state = 0
    
    # Update the actual curve.  Move the vertices on the curve
    # to the new curve points
    if( @vertices.length != @crvpts.length )
        UI.messagebox "Count of curve points is wrong!"
        return
    end

    model = @vertices[0].model
    model.start_operation "Edit Rational Bezier Curve"

    # Move the vertices
    @curve.move_vertices @crvpts
    
    # Update the control points stored with the curve
    @curve.delete_attribute "skp", "crvtype"
    @curve.delete_attribute "skp", "crvpts"
    @curve.set_attribute "jps", "crvtype", "Rational Bezier"
    @curve.set_attribute "jps", "crvpts", @pts
    @curve.set_attribute "jps", "weights",  @weights
    
    model.commit_operation
end

def onMouseMove(flags, x, y, view)
    # Make sure that the control polygon is shown
    view.invalidate if not @drawn
    
    # Move the selected point if state = 1
    if( @state == 1 && @selection )
        @ip.pick view, x, y
        return if not @ip.valid?
        if( @selection >= 0 )
            # Moving a control point
            @pt_to_move = @ip.position
            @pts[@selection] = @pt_to_move
        else
            # moving a segment
            pt = @ip.position
            vec = pt - @pt_to_move
            i = -@selection
            @pts[i-1].offset! vec
            @pts[i].offset! vec
            @pt_to_move = pt
        end
        @crvpts = JPS::BezierCurveManager.points(@pts, @weights, @numseg)
        view.invalidate
    else # state != 1
        # See if we can select something to move
        view.invalidate if( self.pick_point_to_move(x, y, view) )
    end
end

def getMenu(menu)
    menu.add_item("Done") {Sketchup.active_model.select_tool nil}
end

def getExtents
    bb = Geom::BoundingBox.new
    bb.add @pts
    bb
end

def draw(view)
    # Draw the control polygon
    view.drawing_color = "gray"
    view.draw(GL_LINE_STRIP, @pts)
    
    if( @pt_to_move )
        view.draw_points(@pt_to_move, 10, 1, "red");
    end
    
    if( @state == 1 )
        view.drawing_color = "black"
        view.draw(GL_LINE_STRIP, @crvpts)
    end
    
    @drawn = true
end

end # class EditBezierTool

#-----------------------------------------------------------------------------

# Function to test to see if the selection set contains only a Bezier curve
# Returns the curve if there is one or else nil
def self.selected_curve
    ss = Sketchup.active_model.selection
    return nil if not ss.single_object?
    edge = ss.first
    return nil if not edge.kind_of? Sketchup::Edge
    curve = edge.curve
    return edge if not curve
    return nil if (curve.get_attribute("skp", "crvtype") != "Bezier" &&
                   !curve.kind_of?(Sketchup::ArcCurve) &&
                   curve.get_attribute("jps", "crvtype") != "Rational Bezier")
    curve
end

end # module JPS::BezierCurveManager
